import os
from glob import glob
import tarfile
from shutil import copy, copymode


class ConfigMover:
    param_names = ["mode", "perm", "user", "group"]

    def read_rules(self, filelist):
        rules = {}
        with open(filelist, mode='r') as f:
            for line in f:
                stripped = line.strip()
                if stripped.startswith('#') or stripped == '':
                    continue
                new_rule = self.create_rule(stripped)
                if new_rule is None:
                    continue
                (new_key, new_value) = new_rule
                if new_key in rules:
                    raise ConfigMover.FileSchemaError('Duplicating rules')
                rules[new_key] = new_value
        return rules

    def create_rule(self, stripped):
        words = stripped.split()
        path = words[0]
        params = {name: value for name, value in zip(self.param_names, words[1:])}
        if 'mode' not in params:
            return None
        return (path, params)

    def find_files(self, rootfs, filelist, filter_modes=None):
        found = set()
        for path, params in self.read_rules(filelist).items():
            if filter_modes is not None:
                if params['mode'] not in filter_modes:
                    continue
            item_path = os.path.join(rootfs, path.strip('/'))
            for expanded in glob(item_path):
                if os.path.isfile(expanded):
                    if expanded in found:
                        raise ConfigMover.FileSchemaError('Multiple file matching')
                    found.add(expanded)

        return found

    def op_save_cfg(self, root, schema, tar_path, **_):
        files = self.find_files(root, schema, filter_modes=('file'))
        with tarfile.open(name=tar_path, mode='w:') as f:
            for item in files:
                f.add(name=item, arcname=os.path.relpath(item, root), recursive=False)

    def op_save_log(self, root, tar_path, **_):
        src = os.path.join(root, self.default_log_dir())

        with tarfile.open(name=tar_path, mode='w:') as f:
            for dirpath, _, files_rel in os.walk(src):
                files_abs = (os.path.join(dirpath, item) for item in files_rel)
                for item in files_abs:
                    f.add(name=item, arcname=os.path.relpath(item, root), recursive=False)

    def op_save_db(self, db_path, bkp_path, **_):
        os.system("sqlite3 {0} '.backup {1}'".format(db_path, bkp_path))

    def op_dbrestore(self, db_path, bkp_path, **_):
        os.system("sqlite3 {0} '.restore {1}'".format(db_path, bkp_path))

    @staticmethod
    def default_dbbackup_path():
        return 'home/spectracom/config/lafayette.bkp'

    @staticmethod
    def default_db_path():
        return 'srv/www2/app/webroot/lafayette'

    @staticmethod
    def default_log_dir():
        return 'home/spectracom/log'

    @staticmethod
    def log_db_tables():
        return {'log_events'}

    @staticmethod
    def discstat_db_tables():
        return {'log_sys_mons', 'log_eth_mons', 'log_oscillators', 'log_gps_statuses'}

    @staticmethod
    def default_discstat_dir():
        return 'home/spectracom/log/discstats'

    @staticmethod
    def default_cfg_dir():
        return 'home/spectracom/default'

    @staticmethod
    def default_ntpstat_dir():
        return 'home/spectracom/log/ntpstats'

    @staticmethod
    def clear_dir(target):
        if os.path.exists(target):
            for name in os.listdir(target):
                abs_path = os.path.join(target, name)
                if os.path.isfile(abs_path):
                    os.remove(abs_path)

    def op_clear_discstats(self, root, **_):
        log_dir = os.path.join(root, self.default_discstat_dir())
        self.clear_dir(log_dir)

        db_path = os.path.join(root, self.default_db_path())
        if os.path.isfile(db_path):
            for tbl_name in self.discstat_db_tables():
                os.system("sqlite3 {0} 'DELETE FROM {1}'".format(db_path, tbl_name))

    def op_load_cfg(self, root, tar_path, **_):
        with tarfile.open(tar_path, 'r') as tar:
            tar.extractall(root)

    def op_restore_default(self, root, **_):
        default_dir = os.path.join(root, self.default_cfg_dir())
        for node_src, _, files in os.walk(default_dir):
            node_dst = node_src.replace(default_dir, root)
            if not os.path.exists(node_dst):
                os.mkdir(node_dst)
                copymode(src=node_src, dst=node_dst)

            for item in files:
                src_file = os.path.join(node_src, item)
                dst_file = os.path.join(node_dst, item)
                if os.path.lexists(dst_file):
                    os.remove(dst_file)
                copy(src=src_file, dst=dst_file, follow_symlinks=False)

    def op_clear_ntpstats(self, root, **_):
        log_dir = os.path.join(root, self.default_ntpstat_dir())
        self.clear_dir(log_dir)

    class FileSchemaError(ValueError):
        pass
