import re


class UsersService:
    class FileType():
        PASSWORD = '/etc/passwd'
        SHADOW = '/etc/shadow'
        GROUP = '/etc/group'

    @staticmethod
    def __get_file_path(rootfs, file_type):
        return rootfs + file_type

    @staticmethod
    def __read_lines(file):
        return [line.strip() for line in file.readlines()]

    @staticmethod
    def __override_data(file, data):
        file.seek(0)
        file.write("\n".join(data))
        file.truncate()

    @staticmethod
    def get_user_names_by_uid(rootfs, uid):
        result = []
        path = UsersService.__get_file_path(rootfs, UsersService.FileType.PASSWORD)

        with open(path, 'r') as file:
            data = UsersService.__read_lines(file)

        for user_id in uid:
            regex = re.compile(":{0}:[0-9]*:".format(user_id))
            lines = list(filter(lambda line: regex.search(line), data))
            if len(lines) == 0:
                result.append(None)
            elif len(lines) == 1:
                result.append(lines[0].split(':')[0])
            else:
                raise Exception("Detected multiple users with the same UID ({0}) in: {1}".format(id, path))

        return result

    @staticmethod
    def get_users_id_by_gid(rootfs, gid):
        result = []
        path = UsersService.__get_file_path(rootfs, UsersService.FileType.PASSWORD)

        with open(path, 'r') as file:
            data = UsersService.__read_lines(file)

        regex = re.compile(":[0-9]*:{0}:".format(gid))
        lines = list(filter(lambda line: regex.search(line), data))

        for line in lines:
            result.append(int(line.split(':')[2]))

        return result

    @staticmethod
    def is_names_unique(rootfs):
        used_names = []
        with open(UsersService.__get_file_path(rootfs, UsersService.FileType.PASSWORD), 'r') as file:
            data = UsersService.__read_lines(file)

        for line in data:
            name = line.split(':')[0]
            if name in used_names:
                return False
            used_names.append(name)

        return True

    @staticmethod
    def update_passwd(rootfs, new_name, uid):
        with open(UsersService.__get_file_path(rootfs, UsersService.FileType.PASSWORD), 'r+') as file:
            data = UsersService.__read_lines(file)

            for name, user_id in zip(new_name, uid):
                if user_id is not None:
                    updated_data = []
                    regex = re.compile(":{0}:[0-9]*:".format(user_id))
                    for line in data:
                        if name is not None:
                            updated_data.append(re.sub(r"^\w*", name, line) if regex.search(line) else line)
                        else:  # user was removed
                            if not regex.search(line):
                                updated_data.append(line)
                    data = updated_data
            UsersService.__override_data(file, data)

    @staticmethod
    def get_passwords(rootfs, user_names):
        passwords = [None] * len(user_names)

        with open(UsersService.__get_file_path(rootfs, UsersService.FileType.SHADOW), 'r') as file:
            data = UsersService.__read_lines(file)

        for line in data:
            fields = line.split(':')
            name = fields[0]
            password = fields[1]

            for i, user_name in enumerate(user_names):
                if user_name == name:
                    passwords[i] = password

        return passwords

    @staticmethod
    def update_shadow(rootfs, old_names, new_names, passwords):
        with open(UsersService.__get_file_path(rootfs, UsersService.FileType.SHADOW), 'r+') as file:
            data = UsersService.__read_lines(file)

            updated_data = []
            for line in data:
                fields = line.split(':')
                user_was_removed = False
                for old_name, new_name, password in zip(old_names, new_names, passwords):
                    if old_name is not None and fields[0] == old_name:
                        if new_name is not None and password is not None:
                            fields[0] = new_name
                            fields[1] = password
                        elif new_name is None:  # user was removed
                            user_was_removed = True
                        break
                if not user_was_removed:
                    updated_data.append(":".join(fields))

            UsersService.__override_data(file, updated_data)

    @staticmethod
    def update_group(rootfs, current_names, new_names):
        with open(UsersService.__get_file_path(rootfs, UsersService.FileType.GROUP), 'r+') as file:
            data = UsersService.__read_lines(file)

            updated_data = []
            for line in data:
                fields = line.split(':')
                names = fields[3].split(',')

                for old_name, new_name in zip(current_names, new_names):
                    names = [x for x in names if not (x == old_name and new_name is None)]

                for i, name in enumerate(names):
                    for current_name, new_name in zip(current_names, new_names):
                        if name == current_name:
                            names[i] = new_name
                            break

                fields[3] = ",".join(names)
                updated_data.append(":".join(fields))

            UsersService.__override_data(file, updated_data)

    @staticmethod
    def __get_record_by_user_name(file, name):
        with open(file, 'r') as file:
            data = UsersService.__read_lines(file)

        regex = re.compile("^{0}:".format(name))
        lines = list(filter(lambda line: regex.search(line), data))

        if len(lines) == 0:
            raise Exception("No records about user ({0}) in ({1})".format(name, file))
        elif len(lines) == 1:
            return lines[0]
        else:
            raise Exception("Multiple records about user ({0}) in ({1})".format(name, file))

    @staticmethod
    def read_record_from_passwd(rootfs, name):
        file = UsersService.__get_file_path(rootfs, UsersService.FileType.PASSWORD)
        return UsersService.__get_record_by_user_name(file, name)

    @staticmethod
    def read_record_from_shadow(rootfs, name):
        file = UsersService.__get_file_path(rootfs, UsersService.FileType.SHADOW)
        return UsersService.__get_record_by_user_name(file, name)

    @staticmethod
    def __get_available_uid(lines, expected_uid):
        while True:
            is_duplicate_found = False
            for line in lines:
                regex = re.compile(":{0}:[0-9]*:".format(expected_uid))
                if regex.search(line):
                    expected_uid += 1
                    is_duplicate_found = True
            if not is_duplicate_found:
                return expected_uid

    @staticmethod
    def write_record_to_passwd(rootfs, record):
        with open(UsersService.__get_file_path(rootfs, UsersService.FileType.PASSWORD), 'r+') as file:
            data = UsersService.__read_lines(file)

            uid = UsersService.__get_available_uid(data, int(record.split(':')[2]))

            record_fields = record.split(':')
            record_fields[2] = str(uid)
            record = ":".join(record_fields)

            data.append(record)
            UsersService.__override_data(file, data)

    @staticmethod
    def write_record_to_shadow(rootfs, record):
        with open(UsersService.__get_file_path(rootfs, UsersService.FileType.SHADOW), 'r+') as file:
            data = UsersService.__read_lines(file)
            data.append(record)
            UsersService.__override_data(file, data)

    @staticmethod
    def add_user_to_group(rootfs, name, gid):
        with open(UsersService.__get_file_path(rootfs, UsersService.FileType.GROUP), 'r+') as file:
            data = UsersService.__read_lines(file)

            gid_regex = re.compile(":{0}:".format(gid))
            for index, line in enumerate(data):
                if gid_regex.search(line):
                    if name not in line.split(':')[3].split(','):
                        line += ",{0}".format(name)
                        data[index] = line
                    break

            UsersService.__override_data(file, data)

    @staticmethod
    def copy_users(current_rootfs, target_rootfs, names, additional_gid):
        for name in names:
            passwd_record = UsersService.read_record_from_passwd(current_rootfs, name)
            UsersService.write_record_to_passwd(target_rootfs, passwd_record)

            shadow_record = UsersService.read_record_from_shadow(current_rootfs, name)
            UsersService.write_record_to_shadow(target_rootfs, shadow_record)

            UsersService.add_user_to_group(target_rootfs, name, int(passwd_record.split(":")[3]))
            for gid in additional_gid:
                UsersService.add_user_to_group(target_rootfs, name, gid)
